from __future__ import absolute_import
from __future__ import print_function

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import getpass
import os
import sys

from nss.error import NSPRError
import nss.io as io
import nss.nss as nss
import nss.ssl as ssl

# -----------------------------------------------------------------------------

timeout_secs = 3

request = '''\
GET /index.html HTTP/1.0

'''
# -----------------------------------------------------------------------------
# Callback Functions
# -----------------------------------------------------------------------------

def handshake_callback(sock):
    print("-- handshake complete --")
    print("peer: %s" % (sock.get_peer_name()))
    print("negotiated host: %s" % (sock.get_negotiated_host()))
    print()
    print(sock.connection_info_str())
    print("-- handshake complete --")
    print()

def auth_certificate_callback(sock, check_sig, is_server, certdb):
    print("auth_certificate_callback: check_sig=%s is_server=%s" % (check_sig, is_server))
    cert_is_valid = False

    cert = sock.get_peer_certificate()
    pin_args = sock.get_pkcs11_pin_arg()
    if pin_args is None:
        pin_args = ()

    print("cert:\n%s" % cert)

    # Define how the cert is being used based upon the is_server flag.  This may
    # seem backwards, but isn't. If we're a server we're trying to validate a
    # client cert. If we're a client we're trying to validate a server cert.
    if is_server:
        intended_usage = nss.certificateUsageSSLClient
    else:
        intended_usage = nss.certificateUsageSSLServer

    try:
        # If the cert fails validation it will raise an exception, the errno attribute
        # will be set to the error code matching the reason why the validation failed
        # and the strerror attribute will contain a string describing the reason.
        approved_usage = cert.verify_now(certdb, check_sig, intended_usage, *pin_args)
    except Exception as e:
        print(e)
        cert_is_valid = False
        print("Returning cert_is_valid = %s" % cert_is_valid)
        return cert_is_valid

    print("approved_usage = %s" % nss.cert_usage_flags(approved_usage))

    # Is the intended usage a proper subset of the approved usage
    if approved_usage & intended_usage:
        cert_is_valid = True
    else:
        cert_is_valid = False

    # If this is a server, we're finished
    if is_server or not cert_is_valid:
        print("Returning cert_is_valid = %s" % cert_is_valid)
        return cert_is_valid

    # Certificate is OK.  Since this is the client side of an SSL
    # connection, we need to verify that the name field in the cert
    # matches the desired hostname.  This is our defense against
    # man-in-the-middle attacks.

    hostname = sock.get_hostname()
    print("verifying socket hostname (%s) matches cert subject (%s)" % (hostname, cert.subject))
    try:
        # If the cert fails validation it will raise an exception
        cert_is_valid = cert.verify_hostname(hostname)
    except Exception as e:
        print(e)
        cert_is_valid = False
        print("Returning cert_is_valid = %s" % cert_is_valid)
        return cert_is_valid

    print("Returning cert_is_valid = %s" % cert_is_valid)
    return cert_is_valid

# -----------------------------------------------------------------------------
# Client Implementation
# -----------------------------------------------------------------------------

def client():
    valid_addr = False
    # Get the IP Address of our server
    try:
        addr_info = io.AddrInfo(options.hostname)
    except:
        print("ERROR: could not resolve hostname \"%s\"" % options.hostname)
        return

    for net_addr in addr_info:
        net_addr.port = options.port
        sock = ssl.SSLSocket(net_addr.family)
        # Set client SSL socket options
        sock.set_ssl_option(ssl.SSL_SECURITY, True)
        sock.set_ssl_option(ssl.SSL_HANDSHAKE_AS_CLIENT, True)
        sock.set_hostname(options.hostname)

        # Provide a callback which notifies us when the SSL handshake is
        # complete
        sock.set_handshake_callback(handshake_callback)

        # Provide a callback to verify the servers certificate
        sock.set_auth_certificate_callback(auth_certificate_callback,
                                           nss.get_default_certdb())

        try:
            print("try connecting to: %s" % (net_addr))
            sock.connect(net_addr, timeout=io.seconds_to_interval(timeout_secs))
            print("connected to: %s" % (net_addr))
            valid_addr = True
            break
        except:
            continue

    if not valid_addr:
        print("ERROR: could not connect to \"%s\"" % options.hostname)
        return

    try:
        # Talk to the server
        n_received = 0
        sock.send(request.encode('utf-8'))
        while True:
            buf = sock.recv(1024)
            n_received += len(buf)
            if not buf:
                print("\nclient lost connection, received %d bytes" % (n_received))
                break
    except Exception as e:
        print(e)
        sock.shutdown()
        return

    sock.shutdown()
    return

# -----------------------------------------------------------------------------

parser = argparse.ArgumentParser(description='certificate verification example',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('-d', '--db-name',
                    help='NSS database name (e.g. "sql:pki")')

parser.add_argument('-H', '--hostname',
                    help='host to connect to')

parser.add_argument('-p', '--port', type=int,
                    help='host port')

parser.set_defaults(db_name = 'sql:pki',
                    hostname = 'www.verisign.com',
                    port = 443,
                    )

parser.add_argument('--min-ssl-version',
                    help='minimum SSL version')

parser.add_argument('--max-ssl-version',
                    help='minimum SSL version')

options = parser.parse_args()

# Perform basic configuration and setup
try:
    nss.nss_init(options.db_name)
    ssl.set_domestic_policy()

    min_ssl_version, max_ssl_version = \
        ssl.get_supported_ssl_version_range(repr_kind=nss.AsString)
    print("Supported SSL version range: min=%s, max=%s" % \
        (min_ssl_version, max_ssl_version))

    min_ssl_version, max_ssl_version = \
        ssl.get_default_ssl_version_range(repr_kind=nss.AsString)
    print("Default SSL version range: min=%s, max=%s" % \
        (min_ssl_version, max_ssl_version))

    if options.min_ssl_version is not None or \
       options.max_ssl_version is not None:

        if options.min_ssl_version is not None:
            min_ssl_version  = options.min_ssl_version
        if options.max_ssl_version is not None:
            max_ssl_version  = options.max_ssl_version

        print("Setting default SSL version range: min=%s, max=%s" % \
            (min_ssl_version, max_ssl_version))
        ssl.set_default_ssl_version_range(min_ssl_version, max_ssl_version)

        min_ssl_version, max_ssl_version = \
            ssl.get_default_ssl_version_range(repr_kind=nss.AsString)
        print("Default SSL version range now: min=%s, max=%s" % \
            (min_ssl_version, max_ssl_version))

except Exception as e:
    print(str(e), file=sys.stderr)
    sys.exit(1)

client()



from __future__ import absolute_import
from __future__ import print_function

import argparse
import sys

import nss.nss as nss
import nss.error as nss_error
print(sys.path)
import six

#-------------------------------------------------------------------------------

options = None

#-------------------------------------------------------------------------------

def fmt_info(label, item, level=0, hex_data=False):
    fmt_tuples = nss.make_line_fmt_tuples(level, label+':')
    if hex_data:
        fmt_tuples.extend(nss.make_line_fmt_tuples(level+1,
                                                   nss.data_to_hex(item, 16)))
    elif isinstance(item, six.string_types):
        fmt_tuples.extend(nss.make_line_fmt_tuples(level+1, str(item)))
    else:
        fmt_tuples.extend(item.format_lines(level=level+1))
    return nss.indented_format(fmt_tuples)


def generate_key():

    # The AlgorithmID bundles up all the parameters used to
    # generate a symmetric key
    #
    # Note, the defaults for create_pbev2_algorithm_id() are
    # usually correct for most uses, thus one can call
    # create_pbev2_algorithm_id() with no parameters.
    # The parameters are only specified here because this is
    # an example script people can play with.
    alg_id = nss.create_pbev2_algorithm_id(options.pbe_alg,
                                           options.cipher_alg,
                                           options.prf_alg,
                                           options.key_length,
                                           options.iterations,
                                           options.salt)

    if not options.quiet:
        print(fmt_info("create_pbev2_algorithm_id returned()", alg_id))
        print()


    # Pick a PK11 Slot to operate in, we'll use the internal slot
    slot = nss.get_internal_slot()

    # Generate the symmetric key
    sym_key = slot.pbe_key_gen(alg_id, options.password)

    if not options.quiet:
        print(fmt_info("Using password", options.password))
        print()
        print(fmt_info("pbe_key_gen() returned sym_key", sym_key))
        print()

    return alg_id, sym_key

def get_encryption_context(alg_id, sym_key):

    # In order for NSS to encrypt and decrypt data it needs an
    # encryption context to perform the operation in. The cipher used
    # for the encryption context is specified with a PK11 mechanism.
    # The cipher likely also needs additional parameters (i.e. an
    # Initialization Vector (IV) and possibly other values.
    # The get_pbe_crypto_mechanism() call computes the mechanism
    # and parameters for the PBE symmetric key we're using.
    #
    # Because the decryption context needs the same params used in
    # the encryption context we save the param block returned by
    # get_pbe_crypto_mechanism(). So that it can be passed to
    # create_context_by_sym_key() when creating the decryption context.
    # It's often the case the decryption is performed by a separate
    # process so in this example we illustrate exchanging the param
    # as base64 data.

    mechanism, params = alg_id.get_pbe_crypto_mechanism(sym_key)

    # Format the params binary data into a base64 string.  The zero
    # passed for the chars_per_line parameter indicates we want the
    # base64 data as one single string as opposed to a list of wrapped
    # strings.
    params_base64 = params.to_base64(0)

    if not options.quiet:
        print(fmt_info("get_pbe_crypto_mechanism (encrypting) returned mechanism:",
                       nss.key_mechanism_type_name(mechanism)))
        print(fmt_info("get_pbe_crypto_mechanism (encrypting) returned params:",
                       params))
        print()

    # Now we have enough information to create an encrypting context
    # and decrypting the data.

    encrypt_ctx = nss.create_context_by_sym_key(mechanism, nss.CKA_ENCRYPT,
                                                sym_key, params)

    # Return the encrypting context and it's parameter block so that the
    # decryption context can use the same parameter block.
    return encrypt_ctx, params_base64

def get_decryption_context(alg_id, sym_key, params_base64):

    # Build a decryption context using the same parameters used
    # when the encryption context was created.

    # Do NOT use the params returned by get_pbe_crypto_mechanism()
    # because the params often include an IV (Initialization Vector)
    # created with random data, therefore the params used in the
    # encryption context will not match the params needed for the
    # decryption context. Instead use the params used in the
    # encryption context. For interoperability reasons we exchange the
    # params as base64 encoded binary data.

    mechanism, params = alg_id.get_pbe_crypto_mechanism(sym_key)

    # Recreate the params used during encryption by initializing a
    # SecItem from base64 text data (indicated by ascii=True)
    params = nss.SecItem(params_base64, ascii=True)

    if not options.quiet:
        print(fmt_info("get_pbe_crypto_mechanism (decrypting) returned mechanism:",
                       nss.key_mechanism_type_name(mechanism)))
        print()

    # Now we have enough information to create a decrypting context

    decrypt_ctx = nss.create_context_by_sym_key(mechanism, nss.CKA_DECRYPT,
                                                sym_key, params)

    return decrypt_ctx

def do_pbkdf2():

    # Generate a symmetric key
    alg_id, sym_key = generate_key()

    # Get encryption contexts to encrypt
    encrypt_ctx, params_base64 = get_encryption_context(alg_id, sym_key)

    # First encrypt the plain text input

    print(fmt_info("Plain Text", options.plain_text))
    print()

    # Encode the plain text by feeding it to cipher_op getting cipher text back.
    # Append the final bit of cipher text by calling digest_final
    cipher_text = encrypt_ctx.cipher_op(options.plain_text.encode('utf-8'))
    cipher_text += encrypt_ctx.digest_final()

    print(fmt_info("Cipher Text", cipher_text, hex_data=True))
    print()

    # Get decryption contexts to decrypt
    decrypt_ctx = get_decryption_context(alg_id, sym_key, params_base64)

    # Decode the cipher text by feeding it to cipher_op getting plain text back.
    # Append the final bit of plain text by calling digest_final
    decoded_text = decrypt_ctx.cipher_op(cipher_text)
    decoded_text += decrypt_ctx.digest_final()
    decoded_text = decoded_text.decode('utf-8')

    print(fmt_info("Decoded Text", decoded_text))
    print()

#-------------------------------------------------------------------------------
def main():
    global options

    parser = argparse.ArgumentParser(description='Password Based Encryption Example')

    parser.add_argument('-q', '--quiet', action='store_true',
                        help='stiffle chatty output')

    # === NSS Database Group ===
    group = parser.add_argument_group('PBKDF2',
                                      'Specify the PBKDF2 parameters')

    group.add_argument('--pbe-alg',
                       help='password based encryption algorithm')
    group.add_argument('--cipher-alg',
                       help='cipher algorithm')
    group.add_argument('--prf-alg',
                       help='pseudo-random function algorithm')
    group.add_argument('-l', '--key-length',
                       help='number of octets in derived key')
    group.add_argument('-s', '--salt',
                       help='salt as a string, if none then use random data')


    group = parser.add_argument_group('Encryption',
                                      'Specify the encryption parameters')

    group.add_argument('-p', '--password',
                       help='PBE password')

    group.add_argument('-t', '--plain-text',
                       help='string to encrypt')

    parser.set_defaults(pbe_alg = 'SEC_OID_PKCS5_PBKDF2',
                        cipher_alg = 'SEC_OID_AES_256_CBC',
                        prf_alg = 'SEC_OID_HMAC_SHA1',
                        iterations = 100,
                        key_length = 0,
                        salt = None,
                        password = 'password',
                        plain_text = 'Black holes are where God divided by zero - Steven Wright',
                        )

    options = parser.parse_args()

    # Initialize NSS.
    nss.nss_init_nodb()

    do_pbkdf2()

#-------------------------------------------------------------------------------

if __name__ == "__main__":
    sys.exit(main())
